////////////////////////////////////////////////////////////////////////
//
//        File  Name : irc-server.cc
//        Class Name : 
//        Author     : Young-ki Hwang
//        Date       : April 30, 1996
//        Description: Main program of the server part.
//                     
//
///////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <iostream.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include "win.h"
#include "server.h"
#include "member.h"
#include "fork.h"
#include "string.h"



extern "C" sleep(); 
help_handler(server& server1);

main(int argc, char *argv[])
{
  char buf[MAX_MSG_SIZE], buf1[MAX_MSG_SIZE], nickname[MAX_MSG_SIZE],  nick[MAX_MSG_SIZE]="[",
       temp1[MAX_MSG_SIZE],temp2[MAX_MSG_SIZE];
  int pid, socketfd, i, j;

  if(argc != 2)
     { cout << " Usage : irc-server  server_address ";
       exit(1);
     }

  members member;
  forks   fork1;
  pipes   pipe1, pipe2;
  strings string1; 

  server server1(argv[1]); 
  server1.initialize();
  server1.ready_connection();
  
  // Run forever.
  for( ; ; )
   {
    // Check if connection request is from the client. If yes, accept it.
    if( server1.make_connection() )
      {  
       // Notify to new member that how many people has joined.
       for(i=0; i<member.get_numofmember(); i++)
	 { member.get_nickname(i, buf1);
	   while(!server1.send(server1.get_newsockfd(), buf1));
	 }
       

       while(!server1.send(server1.get_newsockfd(), "end"));
       
       // If new member is  going to join, go ahead.
       while(!server1.receive(server1.get_newsockfd(), buf1));
       if( string1.compare( buf1, "ok") )
        {   
	  
	  // Receive a id of the client.
	  while(!server1.receive(server1.get_newsockfd(), nickname));
	  
	  // pipe descriptor between parent and child.
	  pipe1.make_pipe();
	  pipe2.make_pipe();
	  
	  // Insert a new member into the tabel for members.
	  while(! member.insert(server1, pipe1, pipe2,  nickname))
	    while(!server1.receive(server1.get_newsockfd(), nickname));
	  
	  // Make a child.
	  pid = fork1.make_child();
	  
	  // Child process.
	  if( pid == 0 )
	    { 
	      server1.close_oldsocket();
	      pipe1.close_writepipe();
	      pipe2.close_readpipe();
	      
	      string1.concatenate(nick, nickname);
	      string1.concatenate(nick, "]");
	      
	      while(1)
		{ // Receive a message from the client.
		  while(!server1.receive(server1.get_newsockfd(),buf));
		  
		  // If message begins with '/' character.
		  if( buf[0] == '/' )
		    {  string1.scan(buf, temp1);
		       
		       // If someone is going to leave out, 
		       if( string1.compare(temp1, "/quit") )
			 { while(!pipe2.send(buf, nickname));
			   exit(0);
			 }/* end if */
		       
		       // If somebody is going to change his id.
		       else if( string1.compare(temp1, "/nick") )
			 { string1.leftshift(buf, string1.get_length(temp1)+1);
			   string1.scan(buf, nickname);
			   
			   while(!pipe2.send(temp1, nickname));
			   
			   string1.copy(nick, "[");
			   string1.concatenate(nick, nickname);
			   string1.concatenate(nick, "]");
			 }/* end else if */
	      
	      
                       // If someone wants a help menu.
		       else if( string1.compare( temp1, "/help") )
			  help_handler(server1);
			
  
		       // If message is private one to somebody, 
		       else
			 { string1.copy(buf1, nick);
			   string1.concatenate(buf1, " ## private message ## ");
			   string1.leftshift(buf, string1.get_length(temp1));
			   string1.concatenate(buf1, buf);
			
			   string1.leftshift(temp1, 1);
                          for( i=0; i<5; i++ ) 
			    {  string1.scan(temp1, temp2, '/');
                               if(string1.compare(temp2, ""))
                                  break;

			       while(!pipe2.send(buf1, temp2));
                               string1.leftshift(temp1, string1.get_length(temp2));
                               if(temp1[0] == '/');
				 string1.leftshift(temp1, 1);
			    } 

			 }/* end else */

		     }/* end if */
		  
		  
		  // Otherwise: it's a normal message.
		  else 
		    { string1.copy(buf1, nick);
		      string1.concatenate(buf1, buf);
		      while(!pipe2.send(buf1));
		    }
			  
		}/* end while */
	    }/* end if : child*/
	  
	  pipe2.close_writepipe();
	  pipe1.close_readpipe();
	
	}/* end if : string is "ok"*/
      } /* end if : make_connection()*/


    // Check if there is any message from the childs. If yes, read it.
    for(i=0; i<member.get_numofmember(); i++)
      {
	// Read  a message from the pipe.
        if( pipe2.receive(member.get_readpipe(i), buf, nickname) )
	  { 
	   // If someone is leaving, delete the member from the table.
           if( string1.compare(buf, "/quit") )
              member.remove(server1, nickname);
	   
	   // If someone wants to change his nickname, chaange it.
           else if( string1.compare(buf, "/nick") )
	     member.change_nickname(server1, i, nickname);

	   
           // If someone is sending a message to everybody, send it to everyone.
           else if( string1.compare(nickname, "everyone") )
	     {  for(j=0; j<member.get_numofmember(); j++)
		server1.send(member.get_sockfd(j), buf);
	      }/* end else if */

           // If someone is sending a private message, send it to specific someone.
	   else
	     {  if( member.check_exist(nickname) )
	          { server1.send(member.get_sockfd(nickname), buf);
		    server1.send(member.get_sockfd(i), buf);
		   }
	      }/* end else */

	  }/* end if */

      }/* end for : checking from child*/

  
    }/* end for : infinite loop */

}/* end main() */
        
 

//
// help_handler()
// show the help menu to the user.
//
help_handler(server& server1)
{
  int i=server1.get_newsockfd();
  server1.send(i,"========================================================================");
  server1.send(i,"     COMMAND             USAGE                                         ");
  server1.send(i,"========================================================================");
  server1.send(i," /help                  : help menu                                     ");
  server1.send(i," /nick  new_name        : change the id to new_name                     ");
  server1.send(i," /id  message           : send the message only to the person who is id "); 
  server1.send(i," /id1/... /idN  message : send the message to N peoples who are id1,    ");
  server1.send(i,"                          id2, id3, ... , idN                           ");
  server1.send(i," /quit                  : exit the program                              ");
  server1.send(i,"========================================================================");
}


